# Copyright 2012-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import sys
import textwrap
import pytest
import portage
from portage import os
from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs
from portage._global_updates import _do_global_updates
from portage.output import colorize


class MoveEntTestCase(TestCase):
    def testMoveEnt(self):
        ebuilds = {
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "4",
                "SLOT": "2",
            },
        }

        installed = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "4",
            },
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "4",
                "SLOT": "2",
            },
        }

        binpkgs = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "4",
            },
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "4",
                "SLOT": "2",
            },
        }

        updates = textwrap.dedent(
            """
			move dev-libs/A dev-libs/A-moved
		"""
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    user_config={
                        "make.conf": (
                            f'BINPKG_FORMAT="{binpkg_format}"',
                            'FEATURES="-binpkg-signing"',
                        ),
                    },
                )

                settings = playground.settings
                trees = playground.trees
                eroot = settings["EROOT"]
                test_repo_location = settings.repositories["test_repo"].location
                portdb = trees[eroot]["porttree"].dbapi
                vardb = trees[eroot]["vartree"].dbapi
                bindb = trees[eroot]["bintree"].dbapi

                updates_dir = os.path.join(test_repo_location, "profiles", "updates")

                try:
                    ensure_dirs(updates_dir)
                    with open(os.path.join(updates_dir, "1Q-2010"), "w") as f:
                        f.write(updates)

                    # Create an empty updates directory, so that this
                    # repo doesn't inherit updates from the main repo.
                    ensure_dirs(
                        os.path.join(
                            portdb.getRepositoryPath("dont_apply_updates"),
                            "profiles",
                            "updates",
                        )
                    )

                    global_noiselimit = portage.util.noiselimit
                    portage.util.noiselimit = -2
                    try:
                        _do_global_updates(trees._running_eroot, trees, {})
                    finally:
                        portage.util.noiselimit = global_noiselimit

                    # Workaround for cache validation not working
                    # correctly when filesystem has timestamp precision
                    # of 1 second.
                    vardb._clear_cache()

                    # A -> A-moved
                    self.assertRaises(KeyError, vardb.aux_get, "dev-libs/A-1", ["EAPI"])
                    vardb.aux_get("dev-libs/A-moved-1", ["EAPI"])
                    # The original package should still exist because a binary
                    # package move is a copy on write operation.
                    bindb.aux_get("dev-libs/A-1", ["EAPI"])
                    bindb.aux_get("dev-libs/A-moved-1", ["EAPI"])

                    # dont_apply_updates
                    self.assertRaises(
                        KeyError, vardb.aux_get, "dev-libs/A-moved-2", ["EAPI"]
                    )
                    vardb.aux_get("dev-libs/A-2", ["EAPI"])
                    self.assertRaises(
                        KeyError, bindb.aux_get, "dev-libs/A-moved-2", ["EAPI"]
                    )
                    bindb.aux_get("dev-libs/A-2", ["EAPI"])

                finally:
                    playground.cleanup()

    def testMoveEntWithSignature(self):
        ebuilds = {
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "4",
                "SLOT": "2",
            },
        }

        installed = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "4",
            },
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "4",
                "SLOT": "2",
            },
        }

        binpkgs = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "4",
            },
            "dev-libs/A-2::dont_apply_updates": {
                "EAPI": "4",
                "SLOT": "2",
            },
        }

        updates = textwrap.dedent(
            """
			move dev-libs/A dev-libs/A-moved
		"""
        )

        for binpkg_format in ("gpkg",):
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    user_config={
                        "make.conf": (f'BINPKG_FORMAT="{binpkg_format}"',),
                    },
                )

                settings = playground.settings
                trees = playground.trees
                eroot = settings["EROOT"]
                test_repo_location = settings.repositories["test_repo"].location
                portdb = trees[eroot]["porttree"].dbapi
                vardb = trees[eroot]["vartree"].dbapi
                bindb = trees[eroot]["bintree"].dbapi

                updates_dir = os.path.join(test_repo_location, "profiles", "updates")

                try:
                    ensure_dirs(updates_dir)
                    with open(os.path.join(updates_dir, "1Q-2010"), "w") as f:
                        f.write(updates)

                    # Create an empty updates directory, so that this
                    # repo doesn't inherit updates from the main repo.
                    ensure_dirs(
                        os.path.join(
                            portdb.getRepositoryPath("dont_apply_updates"),
                            "profiles",
                            "updates",
                        )
                    )

                    global_noiselimit = portage.util.noiselimit
                    portage.util.noiselimit = -2
                    try:
                        _do_global_updates(trees._running_eroot, trees, {})
                    finally:
                        portage.util.noiselimit = global_noiselimit

                    # Workaround for cache validation not working
                    # correctly when filesystem has timestamp precision
                    # of 1 second.
                    vardb._clear_cache()

                    # A -> A-moved
                    self.assertRaises(KeyError, vardb.aux_get, "dev-libs/A-1", ["EAPI"])
                    vardb.aux_get("dev-libs/A-moved-1", ["EAPI"])
                    # The original package should still exist because a binary
                    # package move is a copy on write operation.
                    bindb.aux_get("dev-libs/A-1", ["EAPI"])
                    print(bindb.aux_get("dev-libs/A-1", "PF"))
                    self.assertRaises(
                        KeyError, bindb.aux_get, "dev-libs/A-moved-1", ["EAPI"]
                    )

                    # dont_apply_updates
                    self.assertRaises(
                        KeyError, vardb.aux_get, "dev-libs/A-moved-2", ["EAPI"]
                    )
                    vardb.aux_get("dev-libs/A-2", ["EAPI"])
                    self.assertRaises(
                        KeyError, bindb.aux_get, "dev-libs/A-moved-2", ["EAPI"]
                    )
                    bindb.aux_get("dev-libs/A-2", ["EAPI"])

                finally:
                    playground.cleanup()

    # Ignore "The loop argument is deprecated" since this argument is conditionally
    # added to asyncio.Lock as needed for compatibility with python 3.9.
    @pytest.mark.filterwarnings("ignore:The loop argument is deprecated")
    @pytest.mark.filterwarnings("error")
    def testMoveEntWithCorruptIndex(self):
        """
        Test handling of the Packages index being stale (bug #920828)
        and gpkg's binpkg-multi-instance handling.

        We expect a UserWarning to be thrown if the gpkg structure is broken,
        so we promote that to an error.
        """
        ebuilds = {
            "dev-libs/A-moved-1::test_repo": {
                "EAPI": "4",
                "SLOT": "2",
            },
            "dev-libs/B-1::test_repo": {"EAPI": "4", "RDEPEND": "dev-libs/A-moved"},
        }

        installed = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "4",
            },
            "dev-libs/B-1::test_repo": {"EAPI": "4", "RDEPEND": "dev-libs/A"},
        }

        binpkgs = {
            "dev-libs/A-1::test_repo": {
                "EAPI": "4",
                "BUILD_ID": "1",
            },
            "dev-libs/B-1::test_repo": {
                "EAPI": "4",
                "BUILD_ID": "1",
                "RDEPEND": "dev-libs/A",
            },
        }

        updates = textwrap.dedent(
            """
			move dev-libs/A dev-libs/A-moved
		"""
        )

        for binpkg_format in ("gpkg",):
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    ebuilds=ebuilds,
                    installed=installed,
                    user_config={
                        "make.conf": (
                            f'BINPKG_FORMAT="{binpkg_format}"',
                            'FEATURES="binpkg-multi-instance pkgdir-index-trusted"',
                        ),
                    },
                    debug=False,
                )

                settings = playground.settings
                trees = playground.trees
                eroot = settings["EROOT"]
                test_repo_location = settings.repositories["test_repo"].location
                portdb = trees[eroot]["porttree"].dbapi
                vardb = trees[eroot]["vartree"].dbapi
                bindb = trees[eroot]["bintree"].dbapi

                updates_dir = os.path.join(test_repo_location, "profiles", "updates")

                try:
                    ensure_dirs(updates_dir)
                    with open(os.path.join(updates_dir, "1Q-2010"), "w") as f:
                        f.write(updates)

                    # Make the Packages index out-of-date
                    os.remove(
                        os.path.join(
                            bindb.bintree.pkgdir, "dev-libs", "A", "A-1-1.gpkg.tar"
                        )
                    )

                    global_noiselimit = portage.util.noiselimit
                    portage.util.noiselimit = -2
                    try:
                        _do_global_updates(trees._running_eroot, trees, {})
                    finally:
                        portage.util.noiselimit = global_noiselimit
                finally:
                    playground.cleanup()



# Copyright 2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os
import subprocess
import sys
import time

import portage
from portage.tests import TestCase, CommandStep, FunctionStep
from portage.tests.resolver.ResolverPlayground import ResolverPlayground


class EmainBinhostTestCase(TestCase):
    def testCompressedIndex(self):
        debug = False

        user_config = {"make.conf": ('FEATURES="-compress-index"',)}

        binpkgs = {
            "app-misc/A-1": {
                "EAPI": "8",
                "DEPEND": "app-misc/B",
                "RDEPEND": "app-misc/C",
            },
        }

        playground = ResolverPlayground(
            binpkgs=binpkgs,
            user_config=user_config,
            debug=debug,
        )
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        bintree = playground.trees[eroot]["bintree"]

        cmds = {}
        for cmd in ("emaint",):
            for bindir in (self.bindir, self.sbindir):
                path = os.path.join(str(bindir), cmd)
                if os.path.exists(path):
                    cmds[cmd] = (portage._python_interpreter, "-b", "-Wd", path)
                    break
            else:
                raise AssertionError(
                    f"{cmd} binary not found in {self.bindir} or {self.sbindir}"
                )

        env = settings.environ()
        env.update(
            {
                "PORTAGE_OVERRIDE_EPREFIX": eprefix,
                "HOME": eprefix,
                "PYTHONDONTWRITEBYTECODE": os.environ.get(
                    "PYTHONDONTWRITEBYTECODE", ""
                ),
            }
        )

        def current_time(offset=0):
            t = time.time() + offset
            return (t, t)

        steps = (
            FunctionStep(
                function=lambda i: self.assertTrue(
                    os.path.exists(bintree._pkgindex_file), f"step {i}"
                ),
            ),
            # The compressed index should not exist yet because compress-index is disabled in make.conf.
            FunctionStep(
                function=lambda i: self.assertFalse(
                    os.path.exists(bintree._pkgindex_file + ".gz"), f"step {i}"
                )
            ),
            CommandStep(
                returncode=os.EX_OK,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--fix"),
            ),
            CommandStep(
                returncode=os.EX_OK,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--check"),
            ),
            FunctionStep(
                function=lambda i: self.assertTrue(
                    os.path.exists(bintree._pkgindex_file + ".gz"), f"step {i}"
                ),
            ),
            FunctionStep(
                function=lambda i: os.unlink(bintree._pkgindex_file + ".gz"),
            ),
            # It should report an error for a missing Packages.gz here.
            CommandStep(
                returncode=1,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--check"),
            ),
            CommandStep(
                returncode=os.EX_OK,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--fix"),
            ),
            CommandStep(
                returncode=os.EX_OK,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--check"),
            ),
            # Bump the timestamp of Packages so that Packages.gz becomes stale.
            FunctionStep(
                function=lambda i: os.utime(
                    bintree._pkgindex_file, current_time(offset=2)
                ),
            ),
            # It should report an error for stale Packages.gz here.
            CommandStep(
                returncode=1,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--check"),
            ),
            CommandStep(
                returncode=os.EX_OK,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--fix"),
            ),
            CommandStep(
                returncode=os.EX_OK,
                env={"FEATURES": "compress-index"},
                command=cmds["emaint"] + ("binhost", "--check"),
            ),
            # It should delete the unwanted Packages.gz here when compress-index is disabled.
            CommandStep(
                returncode=os.EX_OK,
                env={"FEATURES": "-compress-index"},
                command=cmds["emaint"] + ("binhost", "--fix"),
            ),
            FunctionStep(
                function=lambda i: self.assertFalse(
                    os.path.exists(bintree._pkgindex_file + ".gz"), f"step {i}"
                )
            ),
        )

        try:
            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(steps):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )

        finally:
            playground.cleanup()



# Copyright 2022-2025 Gentoo Authors
# Similar to test_gpkg_gpg.py but
# with full emerge calls to test how we control signature verification.

import portage
import shutil
import sys
import tempfile
import subprocess

from portage import os
from portage import shutil
from portage.const import PORTAGE_PYM_PATH, USER_CONFIG_PATH
from portage.gpg import GPG
from portage.process import find_binary
from portage.tests import TestCase, CommandStep, FunctionStep
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class test_gpkg_gpg_emerge_case(TestCase):
    def test_gpkg_require_signed_repo(self):
        debug = False
        gpg = None
        tmpdir = tempfile.mkdtemp()

        binhost_dir = os.path.join(tmpdir, "binhost-verifyme")
        other_binhost_dir = os.path.join(tmpdir, "binhost-dontverify-me")
        os.mkdir(f"{binhost_dir}-incoming")
        os.mkdir(f"{other_binhost_dir}-incoming")
        os.makedirs(os.path.join(binhost_dir, "app-misc", "hello"))
        os.makedirs(os.path.join(other_binhost_dir, "app-misc", "hello"))

        playground = ResolverPlayground(
            user_config={
                "make.conf": (
                    'FEATURES="${FEATURES} buildpkg -binpkg-signing -binpkg-index-trusted"',
                    'BINPKG_FORMAT="gpkg"',
                    'BINPKG_COMPRESS="none"',
                ),
                "binrepos.conf": (
                    "[test-binhost]",
                    f"sync-uri = file://{binhost_dir}",
                    f"location = {binhost_dir}-incoming",
                    "verify-signature = true",
                    "[test-other-binhost]",
                    f"sync-uri = file://{other_binhost_dir}",
                    f"location = {other_binhost_dir}-incoming",
                    "verify-signature = false",
                ),
            },
            ebuilds={
                "app-misc/hello-1": {
                    "EAPI": 8,
                }
            },
        )

        try:
            settings = playground.settings
            gpg = GPG(settings)
            gpg.unlock()

            portage_python = portage._python_interpreter
            emerge_cmd = (
                portage_python,
                "-b",
                "-Wd",
                os.path.join(str(self.bindir), "emerge"),
            )

            settings = playground.settings
            eprefix = settings["EPREFIX"]
            eroot = settings["EROOT"]
            var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
            user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)

            test_commands = (
                # Create a binpkg. This step should succeed because
                # we aren't requiring just-created local binpkgs
                # to be signed.
                CommandStep(
                    returncode=os.EX_OK,
                    command=emerge_cmd
                    + (
                        "--oneshot",
                        "app-misc/hello",
                    ),
                ),
                # Move it to the 'external' binhost we'll fetch from
                FunctionStep(
                    function=lambda i: self.assertTrue(
                        shutil.move(
                            os.path.join(
                                eroot,
                                "pkgdir",
                                "app-misc",
                                "hello",
                                "hello-1-1.gpkg.tar",
                            ),
                            os.path.join(
                                binhost_dir, "app-misc", "hello", "hello-1-1.gpkg.tar"
                            ),
                        ),
                        f"step {i}",
                    )
                ),
                FunctionStep(
                    function=lambda i: self.assertTrue(
                        shutil.copy(
                            os.path.join(eroot, "pkgdir", "Packages"),
                            os.path.join(binhost_dir, "Packages"),
                        ),
                        f"step {i}",
                    ),
                ),
                # Cleanup the internal PKGDIR where it was
                # originally made, as it's now gone
                CommandStep(
                    returncode=os.EX_OK,
                    command=("emaint",)
                    + (
                        "binhost",
                        "--fix",
                    ),
                ),
                # Try to fetch it from the external binhost upon which
                # verification should fail because it was never signed.
                CommandStep(
                    returncode=not os.EX_OK,
                    command=emerge_cmd
                    + (
                        "--getbinpkgonly",
                        "--verbose",
                        "app-misc/hello",
                    ),
                ),
                # FEATURES="binpkg-ignore-signature" should take precedence
                # over binrepos.conf.
                CommandStep(
                    returncode=os.EX_OK,
                    env={"FEATURES": "binpkg-ignore-signature"},
                    command=emerge_cmd
                    + (
                        "--getbinpkgonly",
                        "--verbose",
                        "app-misc/hello",
                    ),
                ),
                #
                # Test with the other binhost now.
                #
                # Move it to the 'external' binhost we'll fetch from
                FunctionStep(
                    function=lambda i: self.assertTrue(
                        shutil.move(
                            os.path.join(
                                binhost_dir,
                                "app-misc",
                                "hello",
                                "hello-1-1.gpkg.tar",
                            ),
                            os.path.join(
                                other_binhost_dir,
                                "app-misc",
                                "hello",
                                "hello-1-1.gpkg.tar",
                            ),
                        ),
                        f"step {i}",
                    )
                ),
                FunctionStep(
                    function=lambda i: self.assertTrue(
                        shutil.copy(
                            os.path.join(binhost_dir, "Packages"),
                            os.path.join(other_binhost_dir, "Packages"),
                        ),
                        f"step {i}",
                    ),
                ),
                # Cleanup the internal PKGDIR where it was
                # originally made, as it's now gone
                CommandStep(
                    returncode=os.EX_OK,
                    command=("emaint",)
                    + (
                        "binhost",
                        "--fix",
                    ),
                ),
                # FEATURES="binpkg-request-signature" should take precedence
                # over binrepos.conf.
                CommandStep(
                    returncode=not os.EX_OK,
                    env={
                        "FEATURES": "binpkg-request-signature",
                        "PORTAGE_TRUST_HELPER": "true",
                    },
                    command=emerge_cmd
                    + (
                        "--getbinpkgonly",
                        "--verbose",
                        "app-misc/hello",
                    ),
                ),
            )

            fake_bin = os.path.join(eprefix, "bin")
            portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")

            path = settings.get("PATH")
            if path is not None and not path.strip():
                path = None
            if path is None:
                path = ""
            else:
                path = ":" + path
            path = fake_bin + path

            pythonpath = os.environ.get("PYTHONPATH")
            if pythonpath is not None and not pythonpath.strip():
                pythonpath = None
            if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
                pass
            else:
                if pythonpath is None:
                    pythonpath = ""
                else:
                    pythonpath = ":" + pythonpath
                pythonpath = PORTAGE_PYM_PATH + pythonpath

            env = {
                "PORTAGE_OVERRIDE_EPREFIX": eprefix,
                "PATH": path,
                "PORTAGE_PYTHON": portage_python,
                "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
                "PYTHONDONTWRITEBYTECODE": os.environ.get(
                    "PYTHONDONTWRITEBYTECODE", ""
                ),
                "PYTHONPATH": pythonpath,
                "PORTAGE_INST_GID": str(os.getgid()),
                "PORTAGE_INST_UID": str(os.getuid()),
                "FEATURES": "buildpkg",
            }

            dirs = [
                playground.distdir,
                fake_bin,
                portage_tmpdir,
                user_config_dir,
                var_cache_edb,
            ]
            true_symlinks = ["chown", "chgrp"]
            needed_binaries = {
                "true": (find_binary("true"), True),
            }

            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(needed_binaries["true"][0], os.path.join(fake_bin, x))

            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, step in enumerate(test_commands):
                if isinstance(step, FunctionStep):
                    try:
                        step.function(i)
                    except Exception as e:
                        if isinstance(e, AssertionError) and f"step {i}" in str(e):
                            raise
                        raise AssertionError(
                            f"step {i} raised {e.__class__.__name__}"
                        ) from e
                    continue

                proc = subprocess.Popen(
                    step.command,
                    env=dict(env.items(), **(step.env or {})),
                    cwd=step.cwd,
                    stdout=stdout,
                )

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != step.returncode:
                        for line in output:
                            sys.stderr.write(portage._unicode_decode(line))

                self.assertEqual(
                    step.returncode,
                    proc.returncode,
                    f"{step.command} (step {i}) failed with exit code {proc.returncode}",
                )
        finally:
            if gpg is not None:
                gpg.stop()
            shutil.rmtree(tmpdir)
            playground.debug = False
            playground.cleanup()
